from setuptools import Extension
from glob import glob
import sys


ext = Extension('graiax.silkcoder._silkv3',
                sources=[*glob('src/c_silkv3/src/*.c'),
                         "src/c_silkv3/coder.cpp",],
                include_dirs=["src/c_silkv3/interface/"])


if sys.byteorder == "big":
    ext.define_macros.append(("_SYSTEM_IS_BIG_ENDIAN", True))


def build(setup_kwargs):
    """
    This function is mandatory in order to build the extensions.
    """

    setup_kwargs.update(ext_modules=[ext])



from io import BytesIO
from . import decode, encode
from .utils import Codec, CoderError, Codec, choose_encoder, play_audio, issilk, iswave
import argparse
from pathlib import Path

parser = argparse.ArgumentParser(prog="silkcoder", description="silkv3的编解码器（超简单ver.）")

subparsers = parser.add_subparsers()
encode_parser = subparsers.add_parser("encode", help="编码")
encode_parser.add_argument('-i', help="输入文件", required=True)
encode_parser.add_argument('--audio_format', help="音频格式，默认为None")
encode_parser.add_argument('--codec', type=Codec, choices=list(Codec), help="编码器(如果需要) 默认为None")
encode_parser.add_argument('--rate', help="silk码率 默认为None 编码情况下此时编码器将会尝试将码率限制在980kb(若时常在10min内，将严守1Mb线)")
encode_parser.add_argument('-ia', '--ios-adaptive', action='store_true', help="ios适配（控制最高码率在24kbps），默认关闭", default=False)
encode_parser.add_argument('output', help="输出文件名")
encode_parser.add_argument('-ss', type=int, help="开始读取时间,对应ffmpeg/avconc中的ss(只能精确到秒) 默认为0(如t为0则忽略)", default=0)
encode_parser.add_argument('-t', type=int, help="持续读取时间,对应ffmpeg/avconc中的t(只能精确到秒) 默认为0(不剪切)", default=0)
encode_parser.set_defaults(func=encode)

decode_parser = subparsers.add_parser("decode", help="解码")
decode_parser.add_argument('-i', help="输入文件", required=True)
decode_parser.add_argument('--audio-format', help="音频格式，默认为None")
decode_parser.add_argument('--codec', type=Codec, choices=list(Codec), help="解码器(如果需要) 默认为None")
decode_parser.add_argument('--rate', help="输出音频码率，解码情况下则会直接传输给ffmpeg")
decode_parser.add_argument('output', help="输出文件名")
decode_parser.set_defaults(func=decode)

player_parser = subparsers.add_parser("play", help="播放")
player_parser.add_argument('input', help="输入文件")
player_parser.set_defaults(func=play_audio)


if __name__ == "__main__":
    args = parser.parse_args()
    dict_args = vars(args)

    if (func := dict_args.pop("func")) != play_audio:
        input_voice = dict_args.pop("i")
        output_voice = dict_args.pop("output")
        func(input_voice, output_voice, **dict_args)
    else:
        b = Path(dict_args["input"]).read_bytes()
        if issilk(b):
            audio = decode(b, audio_format="wav")
        elif iswave(b):
            audio = b
        else:
            encoder = choose_encoder(b)
            if encoder == Codec.libsndfile:
                import soundfile
                data, samplerate = soundfile.read(BytesIO(b))
                soundfile.write(b2 := BytesIO(), data, samplerate, format="wav")
                audio = b2.getvalue()
            elif encoder == Codec.ffmpeg:
                import subprocess
                from .ffmpeg import ffmpeg_coder
                cmd = [ffmpeg_coder, "-i", "cache:pipe:0", "-"]
                shell = subprocess.Popen(cmd,
                                         stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)
                audio, p_err = shell.communicate(input=b)
                if shell.returncode != 0:
                    raise CoderError(f"ffmpeg error:\n{p_err.decode(errors='ignore')}")
            else:
                raise ValueError("Unsupport Format")
        assert audio is not None
        play_audio(audio)


import os
import subprocess
import sys
import wave
from io import BytesIO
from enum import Enum
from pathlib import Path
from typing import Union, Optional

try:
    import imageio_ffmpeg
    imageio_ffmpeg_exists = True
except ImportError:
    imageio_ffmpeg_exists = False

try:
    import soundfile
    import soxr
    libsndfile_available = True
except ImportError:
    libsndfile_available = False

class ArgTypeMixin(Enum):

    @classmethod
    def argtype(cls, s: str) -> Enum:
        try:
            return cls[s]
        except KeyError as e:
            raise ValueError("Not support Value") from e

    def __str__(self):
        return self.name

class Codec(ArgTypeMixin, Enum):
    wave = 0
    ffmpeg = 1
    libsndfile = 2

def choose_encoder(input_bytes: bytes):
    if iswave(input_bytes):
        return Codec.wave
    elif libsndfile_available and is_libsndfile_supported(input_bytes):
        return Codec.libsndfile
    else:
        # 什么叫做万金油啊（叉腰）
        return Codec.ffmpeg


def choose_decoder(audio_format: str):
    audio_format = audio_format.upper()
    if audio_format == 'WAV':
        return Codec.wave
    elif libsndfile_available and is_libsndfile_supported(audio_format):
        return Codec.libsndfile
    else:
        return Codec.ffmpeg


class CoderError(Exception):
    """所有编码/解码的错误"""
    pass


def input_transform(input_: Union[os.PathLike, str, BytesIO, bytes]) -> bytes:
    if isinstance(input_, (os.PathLike, str)):
        return Path(input_).read_bytes()
    elif isinstance(input_, BytesIO):
        return input_.getvalue()
    elif isinstance(input_, bytes):
        return input_
    else:
        raise ValueError("Unsupport format")


def output_transform(output_: Union[os.PathLike, str, BytesIO, None],
                     data: bytes) -> Optional[bytes]:
    if isinstance(output_, (os.PathLike, str)):
        Path(output_).write_bytes(data)
    elif isinstance(output_, BytesIO):
        output_.write(data)
    elif output_ is None:
        return data
    else:
        raise ValueError("Unsupport format")


def iswave(data: bytes):
    """判断音频是否能通过wave标准库解析"""
    try:
        wave.open(BytesIO(data))
        return True
    except (EOFError, wave.Error):
        return False


def issilk(data: bytes):
    """判断音频是否为silkv3格式"""
    f = data[1:10] if data.startswith(b'\x02') else data[:9]
    return f == b"#!SILK_V3"


def is_libsndfile_supported(data: Union[bytes, str]):
    """判断是否被当前libsndfile所支持
    当传入 bytes 的时候，判断是否能被 libsndfile 解析
    当传入 str 的时候，判断该字符串是否在 available_formats 中"""
    if not libsndfile_available:
        return False
    if isinstance(data, bytes):
        try:
            soundfile.info(BytesIO(data))
            return True
        except RuntimeError:
            return False
    elif isinstance(data, str):
        return data in soundfile.available_formats()


def soxr_available(ffmpeg_path: str):
    p = subprocess.Popen(ffmpeg_path,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         encoding="utf-8")
    return "--enable-libsoxr" in p.communicate()[1]


def which(program):
    """类似于 UNIX 中的 which 命令"""
    # Add .exe program extension for windows support
    if os.name == "nt" and not program.endswith(".exe"):
        program += ".exe"

    envdir_list = [os.curdir] + os.environ["PATH"].split(os.pathsep)

    for envdir in envdir_list:
        program_path = os.path.join(envdir, program)
        if os.path.isfile(program_path) and os.access(program_path, os.X_OK):
            return program_path


def get_ffmpeg():
    """获取本机拥有的编解码器"""
    if which("ffmpeg"):
        return "ffmpeg"
    elif imageio_ffmpeg_exists:
        try:
            return imageio_ffmpeg.get_ffmpeg_exe()
        except RuntimeError:
            Warning("Couldn't find ffmpeg, maybe it'll not work")
    else:
        # 找不到，先警告一波
        Warning("Couldn't find ffmpeg, maybe it'll not work")


def play_audio(source: Union[str, bytes]):
    if sys.platform != "win32":
        raise WindowsError("Only support Windows")

    import winsound
    import msvcrt
    import multiprocessing
    import time

    p = multiprocessing.Process(
        target=winsound.PlaySound,
        args=(source, winsound.SND_FILENAME if isinstance(source, str) else winsound.SND_MEMORY),
    )
    p.start()
    print("请按'q'中断")
    while p.is_alive():
        if msvcrt.kbhit() and msvcrt.getch() in b"qQ": break
        time.sleep(0.1)
    p.terminate()
    p.join()


import asyncio
from email.mime import audio
from io import BytesIO
from typing import Dict, Optional, Union

try:
    import soundfile
    import soxr
    libsndfile_available = True
except ImportError:
    libsndfile_available = False

Num = Union[int, float]

VBR_ENCODING_QUALITY = 0x1300
COMPRESSION_LEVEL = 0x1301


def sndfile_encode(data: bytes, audio_format: Optional[str] = None, ss: Num = 0, t: Num = -1):
    with soundfile.SoundFile(BytesIO(data), 'r', format=audio_format) as f:
        samplerate = f.samplerate
        frame = lambda x: int(x * samplerate)
        pcm = f.read(f._prepare_read(frame(ss), None, frame(t) if t > 0 else -1))

    if len(pcm.shape) > 1 and pcm.shape[1] > 1:
        pcm = pcm.mean(axis=1)
    if samplerate != 24000:
        pcm = soxr.resample(pcm, samplerate, 24000)
    soundfile.write(b := BytesIO(), pcm, 24000, "PCM_16", format="RAW")
    return b.getvalue()


def sndfile_decode(data: bytes,
                   audio_format: str,
                   subtype: Optional[str] = None,
                   quality: Optional[float] = None,
                   metadata: Optional[Dict[str, str]] = None):
    if quality is not None and 0 <= quality <= 1:
        raise ValueError("vbr should between 0 and 1")
    pcm, samplerate = soundfile.read(BytesIO(data),
                                     samplerate=24000,
                                     channels=1,
                                     subtype="PCM_16",
                                     format="RAW")
    with soundfile.SoundFile(b := BytesIO(),
                             'w',
                             samplerate=samplerate,
                             channels=1,
                             format=audio_format,
                             subtype=subtype) as f:
        if metadata:
            for k, v in metadata.items():
                setattr(f, k, v)
        if quality is not None:
            q = soundfile._ffi.new("double*", quality)
            ret = soundfile._snd.sf_command(
                f._file, COMPRESSION_LEVEL if audio_format == "flac" else VBR_ENCODING_QUALITY, q,
                soundfile._ffi.sizeof(q))
            if ret == soundfile._snd.SF_FALSE:
                err = soundfile._snd.sf_error(f._file)
                raise OSError(err, "Error setting quality for the file")
        f.write(pcm)
    return b.getvalue()


async def async_sndfile_encode(data: bytes,
                               audio_format: Optional[str] = None,
                               ss: Num = 0,
                               t: Num = -1):
    return await asyncio.get_running_loop().run_in_executor(None, sndfile_encode, data,
                                                            audio_format, ss, t)


async def async_sndfile_decode(data: bytes,
                               audio_format: str,
                               subtype: Optional[str] = None,
                               quality: Optional[float] = None,
                               metadata: Optional[Dict[str, str]] = None):
    return await asyncio.get_running_loop().run_in_executor(None, sndfile_decode, data,
                                                            audio_format, subtype, quality,
                                                            metadata)


__all__ = [
    "libsndfile_available", "sndfile_encode", "sndfile_decode", "async_sndfile_encode",
    "async_sndfile_decode"
]



from . import _silkv3
import asyncio


def silk_encode(data: bytes, rate: int = -1, tencent: bool = True, ios_adaptive: bool = False):
    if rate < 0:
        #保证压制出来的音频在1000kb上下，若音频时常在10min以内而不超过1Mb
        rate = min(int(980 * 1024 / (len(data) / 24000 / 2) * 8), 24000 if ios_adaptive else 100000)
    return _silkv3.encode(data, rate, tencent)


async def async_silk_encode(data: bytes,
                            rate: int = -1,
                            tencent: bool = True,
                            ios_adaptive: bool = False):
    if rate < 0:
        #保证压制出来的音频在1000kb上下，若音频时常在10min以内而不超过1Mb
        rate = min(int(980 * 1024 / (len(data) / 24000 / 2) * 8), 24000 if ios_adaptive else 100000)
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, _silkv3.encode, data, rate, tencent)


def silk_decode(data: bytes):
    return _silkv3.decode(data)


async def async_silk_decode(data: bytes):
    return await asyncio.get_running_loop().run_in_executor(None, _silkv3.decode, data)


__all__ = ["silk_encode", "silk_decode", "async_silk_encode", "async_silk_decode"]


import audioop
import wave
from io import BytesIO
from typing import Union

Num = Union[int, float]


# 根据实验，audioop 在运行的时候还是占用GIL
# 所以 to_thread 没用，不如直接同步读取
def wav_encode(data: bytes, ss: Num = 0, t: Num = -1):

    with wave.open(BytesIO(data), 'rb') as wav:
        para = wav.getparams()

        rate_trans = lambda x: int(x * para.framerate)
        wav_data = (wav.readframes(rate_trans(ss + t))[rate_trans(ss):]
                    if t > 0 else wav.readframes(wav.getnframes()))

        if para.framerate != 24000:
            wav_data = audioop.ratecv(wav_data, para.sampwidth, para.nchannels, para.framerate,
                                      24000, None)[0]
        if para.nchannels != 1:
            wav_data = audioop.tomono(wav_data, para.sampwidth, 0.5, 0.5)
        if para.sampwidth != 2:
            wav_data = audioop.lin2lin(wav_data, para.sampwidth, 2)
        return wav_data


def wav_decode(data: bytes):
    with wave.open(b := BytesIO(), 'wb') as wav_out:
        wav_out.setnchannels(1)
        wav_out.setsampwidth(2)
        wav_out.setframerate(24000)
        wav_out.writeframes(data)
    return b.getvalue()


__all__ = ["wav_encode", "wav_decode"]



import asyncio
import subprocess
import sys
from typing import Dict, List, Optional, Union

from .utils import CoderError, get_ffmpeg, soxr_available

PIPE = subprocess.PIPE
Num = Union[int, float]

ffmpeg_coder = get_ffmpeg()
if ffmpeg_coder is not None:
    soxr = soxr_available(ffmpeg_coder)
    ffmpeg_available = True
else:
    # 因为不知道到底有没有 ffmpeg，所以默认不使用 soxr
    ffmpeg_coder, soxr = "ffmpeg", False
    ffmpeg_available = False


def get_ffmpeg_encode_cmd(audio_format: Optional[str], ss: Num, t: Num,
                          ffmpeg_para: Optional[List[str]]):
    cmd = [ffmpeg_coder]
    if audio_format is not None: cmd += ['-f', audio_format]

    input_cmd = ["-read_ahead_limit", "-1", "-i", "cache:pipe:0"]

    cmd += [
        '-ss',
        str(ss if isinstance(ss, int) else round(ss, 3)), *input_cmd, '-t',
        str(t if isinstance(t, int) else round(t, 3))
    ] if t > 0 else input_cmd
    if ffmpeg_para: cmd += ffmpeg_para
    if soxr: cmd += ['-af', 'aresample=resampler=soxr']
    cmd += ['-ar', '24000', '-ac', '1', '-y', '-vn', '-loglevel', 'error', '-f', 's16le', '-']
    return cmd


def ffmpeg_encode(data: bytes,
                  audio_format: Optional[str] = None,
                  ss: Num = 0,
                  t: Num = -1,
                  ffmpeg_para: Optional[List[str]] = None):
    cmd = get_ffmpeg_encode_cmd(audio_format, ss, t, ffmpeg_para)
    shell = subprocess.Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    p_out, p_err = shell.communicate(input=data)
    if shell.returncode != 0:
        raise CoderError(f"ffmpeg error:\n{p_err.decode(errors='ignore')}")
    return p_out


async def async_ffmpeg_encode(data: bytes,
                              audio_format: Optional[str] = None,
                              ss: Num = 0,
                              t: Num = -1,
                              ffmpeg_para: Optional[List[str]] = None):
    cmd = get_ffmpeg_encode_cmd(audio_format, ss, t, ffmpeg_para)
    shell = await asyncio.create_subprocess_exec(*cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    p_out, p_err = await shell.communicate(input=data)
    if shell.returncode != 0:
        raise CoderError(f"ffmpeg error:\n{p_err.decode(errors='ignore')}")
    return p_out


def get_ffmpeg_decode_cmd(audio_format: str, ffmpeg_para: Optional[List[str]],
                          rate: Optional[Union[int, str]],
                          metadata: Optional[Dict[str, Union[str, Num]]]):
    cmd = [ffmpeg_coder, '-f', 's16le', '-ar', '24000', '-ac', '1', '-i', 'pipe:']
    if audio_format is not None: cmd += ['-f', audio_format]
    if rate is not None: cmd += ['-b:a', str(rate)]
    if metadata is not None:
        for k, v in metadata.items():
            cmd += ["-metadata", f"{k}={v}"]
    if ffmpeg_para is not None: cmd += [str(a) for a in ffmpeg_para]
    if sys.platform == 'darwin' and audio_format == 'mp3':
        cmd += ["-write_xing", "0"]

    cmd += ['-y', '-loglevel', 'error', 'pipe:']
    return cmd


def ffmpeg_decode(data,
                  audio_format: str,
                  ffmpeg_para: Optional[List[str]] = None,
                  rate: Optional[Union[int, str]] = None,
                  metadata: Optional[Dict[str, Union[str, Num]]] = None):
    cmd = get_ffmpeg_decode_cmd(audio_format, ffmpeg_para, rate, metadata)
    shell = subprocess.Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    p_out, p_err = shell.communicate(input=data)
    if shell.returncode != 0:
        raise CoderError(f"ffmpeg error:\n{p_err.decode(errors='ignore')}")
    return p_out


async def async_ffmpeg_decode(data,
                              audio_format: str,
                              ffmpeg_para: Optional[List[str]] = None,
                              rate: Optional[Union[int, str]] = None,
                              metadata: Optional[Dict[str, Union[str, Num]]] = None):
    cmd = get_ffmpeg_decode_cmd(audio_format, ffmpeg_para, rate, metadata)
    shell = await asyncio.create_subprocess_exec(*cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    p_out, p_err = await shell.communicate(input=data)
    if shell.returncode != 0:
        raise CoderError(f"ffmpeg error:\n{p_err.decode(errors='ignore')}")
    return p_out


__all__ = [
    "ffmpeg_encode", "ffmpeg_decode", "async_ffmpeg_encode", "async_ffmpeg_decode",
    "ffmpeg_available"
]



from typing import Dict, Union, Optional, Literal, List, overload
from io import BytesIO
from os import PathLike
from .utils import Codec
from numbers import Real

filelike = Union[PathLike, str, BytesIO]
Num = Union[int, float]

@overload
async def async_encode(input_voice: Union[filelike, bytes],
                       output_voice: Union[filelike, None] = None,
                       /,
                       codec: Literal[Codec.wave] = Codec.wave,
                       rate: int = -1,
                       ss: Num = 0,
                       t: Num = -1,
                       tencent: bool = True,
                       ios_adaptive: bool = False) -> Optional[bytes]:
    """
    将音频文件转化为 silkv3 格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件
        output_voice(os.PathLike, str, BytesIO, None) 输出文件(silk)，默认为None，为None时将返回bytes

        codec(Codec.wave) 编码器，这里是 python 的 wave 标准库
        rate(int) silk码率 默认为None 此时编码器将会尝试将码率限制在980kb (若时常在10min内，将严守1Mb线)
        ss(Num) 开始读取时间,对应 ffmpeg 中的ss (只能精确到秒) 默认为0(如t为0则忽略)
        t(Num) 持续读取时间,对应 ffmpeg 中的 t (只能精确到秒) 默认为0(不剪切)
        tencent(bool) 是否转化成腾讯的格式
        ios_adaptive(bool) 是否适配 iOS 设备（iOS 的音频码率上限比其他平台低）
    """
    ...

@overload
async def async_encode(input_voice: Union[filelike, bytes],
                       output_voice: Union[filelike, None] = None,
                       /,
                       codec: Literal[Codec.libsndfile] = Codec.libsndfile,
                       audio_format: Optional[str] = None,
                       rate: int = -1,
                       ss: Num = 0,
                       t: Num = -1,
                       tencent: bool = True,
                       ios_adaptive: bool = False) -> Optional[bytes]:
    """
    将音频文件转化为 silkv3 格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件
        output_voice(os.PathLike, str, BytesIO, None) 输出文件(silk)，默认为None，为None时将返回bytes

        codec(Codec.libsndfile) 编码器，这里是 libsndfile
        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由 libsndfile 解析格式)
        rate(int) silk码率 默认为None 此时编码器将会尝试将码率限制在980kb (若时常在10min内，将严守1Mb线)
        ss(Num) 开始读取时间,对应 ffmpeg 中的ss (只能精确到秒) 默认为0(如t为0则忽略)
        t(Num) 持续读取时间,对应 ffmpeg 中的 t (只能精确到秒) 默认为0(不剪切)
        tencent(bool) 是否转化成腾讯的格式
        ios_adaptive(bool) 是否适配 iOS 设备（iOS 的音频码率上限比其他平台低）
    """
    ...

@overload
async def async_encode(input_voice: Union[filelike, bytes],
                       output_voice: Union[filelike, None] = None,
                       /,
                       codec: Literal[Codec.ffmpeg] = Codec.ffmpeg,
                       audio_format: Optional[str] = None,
                       rate: int = -1,
                       ss: Num = 0,
                       t: Num = -1,
                       tencent: bool = True,
                       ios_adaptive: bool = False,
                       ffmpeg_para: Optional[List[str]] = None) -> Optional[bytes]:
    """
    将音频文件转化为 silkv3 格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件
        output_voice(os.PathLike, str, BytesIO, None) 输出文件(silk)，默认为None，为None时将返回bytes

        codec(Codec.ffmpeg) 编码器，这里是 ffmpeg
        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由 ffmpeg 解析格式)
        rate(int) silk码率 默认为None 此时编码器将会尝试将码率限制在980kb (若时常在10min内，将严守1Mb线)
        ss(Num) 开始读取时间,对应 ffmpeg 中的ss (只能精确到秒) 默认为0(如t为0则忽略)
        t(Num) 持续读取时间,对应 ffmpeg 中的 t (只能精确到秒) 默认为0(不剪切)
        tencent(bool) 是否转化成腾讯的格式
        ios_adaptive(bool) 是否适配 iOS 设备（iOS 的音频码率上限比其他平台低）
        ffmpeg_para(list) 额外的 ffmpeg 参数（如 ['-ar', '24000']）
    """
    ...

@overload
async def async_decode(input_voice: Union[filelike, bytes],
                       output_voice: Union[filelike, None] = None,
                       /,
                       codec: Literal[Codec.wave] = Codec.wave) -> Optional[bytes]:
    """
    将silkv3音频转换为其他音频格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件(silk)
        output_voice(os.PathLike, str, BytesIO, None) 输出文件，默认为None，为None时将返回bytes
        codec(Codec.wave) 编码器，这里是 python 的 wave 标准库
    """
    ...

@overload
async def async_decode(input_voice: Union[filelike, bytes],
                       output_voice: Union[filelike, None] = None,
                       /,
                       codec: Literal[Codec.libsndfile] = Codec.libsndfile,
                       audio_format: Optional[str] = None,
                       subtype: Optional[str] = None,
                       quality: Optional[float] = None,
                       metadata: Optional[Dict[str, str]]= None) -> Optional[bytes]:
    """
    将silkv3音频转换为其他音频格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件(silk)
        output_voice(os.PathLike, str, BytesIO, None) 输出文件，默认为None，为None时将返回bytes

        codec(Codec.libsndfile) 编码器，这里是 libsndfile
        audio_format(str) 音频格式(如mp3, ogg) 默认为None（此时将由 libsndfile 解析格式）
        quality(float) 压缩品质，要求在0到1之间
        metadata(dict) 音频标签 默认为 None
    """
    ...

@overload
async def async_decode(input_voice: Union[filelike, bytes],
                       output_voice: Union[filelike, None] = None,
                       /,
                       codec: Literal[Codec.ffmpeg] = Codec.ffmpeg,
                       audio_format: Optional[str] = None,
                       rate: Optional[Union[int, str]] = None,
                       metadata: Optional[Dict[str, str]] = None,
                       ffmpeg_para: Optional[List[str]] = None) -> Optional[bytes]:
    """
    将silkv3音频转换为其他音频格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件(silk)
        output_voice(os.PathLike, str, BytesIO, None) 输出文件，默认为None，为None时将返回bytes

        codec(Codec.ffmpeg) 编码器，这里是 ffmpeg
        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由 ffmpeg 解析格式)
        rate(int) 码率 对应ffmpeg/avconc中"-ab"参数 默认为None
        metadata(dict) 音频标签 将会转化为ffmpeg/avconc参数 如"-metadata title=xxx" 默认为None
        ffmpeg_para(list) 额外的 ffmpeg 参数（如 ['-ar', '24000']）
    """
    ...

@overload
def encode(input_voice: Union[filelike, bytes],
           output_voice: Union[filelike, None] = None,
           /,
           codec: Literal[Codec.wave] = Codec.wave,
           rate: int = -1,
           ss: Num = 0,
           t: Num = -1,
           tencent: bool = True,
           ios_adaptive: bool = False) -> Optional[bytes]:
    """
    将音频文件转化为 silkv3 格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件
        output_voice(os.PathLike, str, BytesIO, None) 输出文件(silk)，默认为None，为None时将返回bytes
        
        codec(Codec.wave) 编码器，这里是 python 的 wave 标准库
        rate(int) silk码率 默认为None 此时编码器将会尝试将码率限制在980kb (若时常在10min内，将严守1Mb线)
        ss(Num) 开始读取时间,对应 ffmpeg 中的ss (只能精确到秒) 默认为0(如t为0则忽略)
        t(Num) 持续读取时间,对应 ffmpeg 中的 t (只能精确到秒) 默认为0(不剪切)
        tencent(bool) 是否转化成腾讯的格式
        ios_adaptive(bool) 是否适配 iOS 设备（iOS 的音频码率上限比其他平台低）
    """
    ...

@overload
def encode(input_voice: Union[filelike, bytes],
           output_voice: Union[filelike, None] = None,
           codec: Literal[Codec.libsndfile] = Codec.libsndfile,
           /,
           audio_format: Optional[str] = None,
           rate: int = -1,
           ss: Num = 0,
           t: Num = -1,
           tencent: bool = True,
           ios_adaptive: bool = False) -> Optional[bytes]:
    """
    将音频文件转化为 silkv3 格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件
        output_voice(os.PathLike, str, BytesIO, None) 输出文件(silk)，默认为None，为None时将返回bytes
        codec(Codec.libsndfile) 编码器，这里是 libsndfile

        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由 libsndfile 解析格式)
        rate(int) silk码率 默认为None 此时编码器将会尝试将码率限制在980kb (若时常在10min内，将严守1Mb线)
        ss(Num) 开始读取时间,对应 ffmpeg 中的ss (只能精确到秒) 默认为0(如t为0则忽略)
        t(Num) 持续读取时间,对应 ffmpeg 中的 t (只能精确到秒) 默认为0(不剪切)
        tencent(bool) 是否转化成腾讯的格式
        ios_adaptive(bool) 是否适配 iOS 设备（iOS 的音频码率上限比其他平台低）
    """
    ...

@overload
def encode(input_voice: Union[filelike, bytes],
           output_voice: Union[filelike, None] = None,
           /,
           codec: Literal[Codec.ffmpeg] = Codec.ffmpeg,
           audio_format: Optional[str] = None,
           rate: int = -1,
           ss: Num = 0,
           t: Num = -1,
           tencent: bool = True,
           ios_adaptive: bool = False,
           ffmpeg_para: Optional[List[str]] = None) -> Optional[bytes]:
    """
    将音频文件转化为 silkv3 格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件
        output_voice(os.PathLike, str, BytesIO, None) 输出文件(silk)，默认为None，为None时将返回bytes

        codec(Codec.ffmpeg) 编码器，这里是 ffmpeg
        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由 ffmpeg 解析格式)
        rate(int) silk码率 默认为None 此时编码器将会尝试将码率限制在980kb (若时常在10min内，将严守1Mb线)
        ss(Num) 开始读取时间,对应 ffmpeg 中的ss (只能精确到秒) 默认为0(如t为0则忽略)
        t(Num) 持续读取时间,对应 ffmpeg 中的 t (只能精确到秒) 默认为0(不剪切)
        tencent(bool) 是否转化成腾讯的格式
        ios_adaptive(bool) 是否适配 iOS 设备（iOS 的音频码率上限比其他平台低）
        ffmpeg_para(list) 额外的 ffmpeg 参数（如 ['-ar', '24000']）
    """
    ...

@overload
def decode(input_voice: Union[filelike, bytes],
           output_voice: Union[filelike, None] = None,
           /,
           codec: Literal[Codec.wave] = Codec.wave) -> Optional[bytes]:
    """
    将silkv3音频转换为其他音频格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件(silk)
        output_voice(os.PathLike, str, BytesIO, None) 输出文件，默认为None，为None时将返回bytes
        codec(Codec.wave) 编码器，这里是 python 的 wave 标准库
    """
    ...

@overload
def decode(input_voice: Union[filelike, bytes],
           output_voice: Union[filelike, None] = None,
           /,
           codec: Literal[Codec.libsndfile] = Codec.libsndfile,
           audio_format: Optional[str] = None,
           quality: Optional[float] = None,
           metadata: Optional[Dict[str, str]] = None) -> Optional[bytes]:
    """
    将silkv3音频转换为其他音频格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件(silk)
        output_voice(os.PathLike, str, BytesIO, None) 输出文件，默认为None，为None时将返回bytes

        codec(Codec.libsndfile) 编码器，这里是 libsndfile
        audio_format(str) 音频格式(如mp3, ogg) 默认为None（此时将由 libsndfile 解析格式）
        quality(float) 压缩品质，要求在0到1之间
        metadata(dict) 音频标签 默认为 None
    """
    ...

@overload
def decode(input_voice: Union[filelike, bytes],
           output_voice: Union[filelike, None] = None,
           /,
           codec: Literal[Codec.ffmpeg] = Codec.ffmpeg,
           audio_format: Optional[str] = None,
           rate: Optional[Union[int, str]] = None,
           metadata: Optional[Dict[str, str]] = None,
           ffmpeg_para: Optional[List[str]] = None) -> Optional[bytes]:
    """
    将silkv3音频转换为其他音频格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件(silk)
        output_voice(os.PathLike, str, BytesIO, None) 输出文件，默认为None，为None时将返回bytes

        codec(Codec.ffmpeg) 编码器，这里是 ffmpeg
        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由 ffmpeg 解析格式)
        rate(int) 码率 对应ffmpeg/avconc中"-ab"参数 默认为None
        metadata(dict) 音频标签 将会转化为ffmpeg/avconc参数 如"-metadata title=xxx" 默认为None
        ffmpeg_para(list) 额外的 ffmpeg 参数（如 ['-ar', '24000']）
    """
    ...


"""
一个不占GIL锁的SilkV3编解码器
注：单个音频压制还是单线程，但是压制时不占用GIL锁
"""
import os
import sys
from io import BytesIO
from pathlib import Path
from typing import Optional, Union

from .ffmpeg import *
from .libsndfile import *
from .utils import Codec, input_transform, output_transform, choose_decoder, choose_encoder
from .wav import *

try:
    from .silkv3 import *
except RuntimeError as e:
    if sys.platform == "win32":
        raise RuntimeError(
            "It seems that you machine doesn't have Visual C++ runtime.\n"
            "You can download it from https://docs.microsoft.com/zh-CN/cpp/windows/latest-supported-vc-redist"
        ) from e

filelike = Union[os.PathLike, str, BytesIO]
Num = Union[int, float]


async def async_encode(input_voice: Union[filelike, bytes],
                       output_voice: Union[filelike, None] = None,
                       /,
                       codec: Optional[Codec] = None,
                       rate: int = -1,
                       ss: Num = 0,
                       t: Num = -1,
                       tencent: bool = True,
                       ios_adaptive: bool = False,
                       **kwargs) -> Optional[bytes]:
    """
    将音频文件转化为 silkv3 格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件
        output_voice(os.PathLike, str, BytesIO, None) 输出文件(silk)，默认为None，为None时将返回bytes
        codec(Codec) 编码器，可选 wave, libsndfile, ffmpeg 默认状态下会让程序自行判断

        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由所选处理器解析格式)
        rate(int) silk码率 默认为None 此时编码器将会尝试将码率限制在980kb (若时常在10min内，将严守1Mb线)
        ss(Num) 开始读取时间,对应 ffmpeg 中的ss (只能精确到秒) 默认为0(如t为0则忽略)
        t(Num) 持续读取时间,对应 ffmpeg 中的 t (只能精确到秒) 默认为0(不剪切)
        tencent(bool) 是否转化成腾讯的格式
        ios_adaptive(bool) 是否适配 iOS 设备（iOS 的音频码率上限比其他平台低）
        ffmpeg_para(list) 额外的 ffmpeg 参数(假设是 ffmpeg 的话)（如 ['-ar', '24000']）
    """
    input_bytes = input_transform(input_voice)
    if codec is None:
        codec = choose_encoder(input_bytes)

    if codec == Codec.wave:
        pcm = wav_encode(input_bytes, ss, t)
    elif codec == Codec.libsndfile:
        audio_format = kwargs.get("audio_format")
        pcm = await async_sndfile_encode(input_bytes, audio_format, ss, t)
    else:
        audio_format = kwargs.get("audio_format")
        ffmpeg_para = kwargs.get("ffmpeg_para")
        pcm = await async_ffmpeg_encode(input_bytes, audio_format, ss, t, ffmpeg_para)

    silk = await async_silk_encode(pcm, rate, tencent, ios_adaptive)

    return output_transform(output_voice, silk)


async def async_decode(input_voice: Union[filelike, bytes],
                       output_voice: Union[filelike, None] = None,
                       /,
                       codec: Optional[Codec] = None,
                       audio_format: Optional[str] = None,
                       **kwargs) -> Optional[bytes]:
    """
    将silkv3音频转换为其他音频格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件(silk)
        output_voice(os.PathLike, str, BytesIO, None) 输出文件，默认为None，为None时将返回bytes
        codec(Codec) 编码器，可选 wave, libsndfile, ffmpeg 默认状态下会让程序自行判断

        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由ffmpeg解析格式)
        rate(int) 码率 对应ffmpeg/avconc中"-ab"参数 默认为None
        metadata(dict) 音频标签
        ffmpeg_para(list) 额外的 ffmpeg 参数(假设是 ffmpeg 的话)（如 ['-ar', '24000']）
    """
    input_bytes = input_transform(input_voice)

    if audio_format is None:
        if isinstance(output_voice, (os.PathLike, str)):
            audio_format = Path(output_voice).suffix[1:]
        else:
            raise ValueError("Pls tell me what audio format to use")
    if codec is None:
        codec = choose_decoder(audio_format)

    pcm = await async_silk_decode(input_bytes)

    if codec == Codec.wave:
        audio = wav_decode(pcm)
    elif codec == Codec.libsndfile:
        metadata = kwargs.get("metadata")
        quality = kwargs.get("quality")
        subtype = kwargs.get("subtype")
        audio = await async_sndfile_decode(pcm, audio_format, subtype, quality, metadata)
    elif codec == Codec.ffmpeg:
        rate = kwargs.get("rate")
        metadata = kwargs.get("metadata")
        ffmpeg_para = kwargs.get("ffmpeg_para")
        audio = await async_ffmpeg_decode(
            pcm,
            audio_format,
            ffmpeg_para,
            rate,
            metadata,
        )

    return output_transform(output_voice, audio)


def encode(input_voice: Union[filelike, bytes],
           output_voice: Union[filelike, None] = None,
           /,
           codec: Optional[Codec] = None,
           rate: int = -1,
           ss: Num = 0,
           t: Num = -1,
           tencent: bool = True,
           ios_adaptive: bool = False,
           **kwargs) -> Optional[bytes]:
    """
    将音频文件转化为 silkv3 格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件
        output_voice(os.PathLike, str, BytesIO, None) 输出文件(silk)，默认为None，为None时将返回bytes
        codec(Codec) 编码器，可选 wave, libsndfile, ffmpeg 默认状态下会让程序自行判断

        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由 ffmpeg 解析格式)
        rate(int) silk码率 默认为None 此时编码器将会尝试将码率限制在980kb (若时常在10min内，将严守1Mb线)
        ss(Num) 开始读取时间,对应 ffmpeg 中的ss (只能精确到秒) 默认为0(如t为0则忽略)
        t(Num) 持续读取时间,对应 ffmpeg 中的 t (只能精确到秒) 默认为0(不剪切)
        tencent(bool) 是否转化成腾讯的格式
        ios_adaptive(bool) 是否适配 iOS 设备（iOS 的音频码率上限比其他平台低）
        ffmpeg_para(list) 额外的 ffmpeg 参数(假设是 ffmpeg 的话)（如 ['-ar', '24000']）
    """

    input_bytes = input_transform(input_voice)

    if codec is None:
        codec = choose_encoder(input_bytes)

    if codec == Codec.wave:
        pcm = wav_encode(input_bytes, ss, t)
    elif codec == Codec.libsndfile:
        audio_format = kwargs.get("audio_format")
        pcm = sndfile_encode(input_bytes, audio_format, ss, t)
    else:
        audio_format = kwargs.get("audio_format")
        ffmpeg_para = kwargs.get("ffmpeg_para")
        pcm = ffmpeg_encode(input_bytes, audio_format, ss, t, ffmpeg_para)

    silk = silk_encode(pcm, rate, tencent, ios_adaptive)
    return output_transform(output_voice, silk)


def decode(input_voice: Union[filelike, bytes],
           output_voice: Union[filelike, None] = None,
           /,
           codec: Optional[Codec] = None,
           audio_format: Optional[str] = None,
           **kwargs) -> Optional[bytes]:
    """
    将silkv3音频转换为其他音频格式

    Args:
        input_voice(os.PathLike, str, BytesIO, bytes) 输入文件(silk)
        output_voice(os.PathLike, str, BytesIO, None) 输出文件，默认为None，为None时将返回bytes
        audio_format(str) 音频格式(如mp3, ogg) 默认为None(此时将由ffmpeg解析格式)
        codec(str) 编码器(如果需要) 默认为None
        ensure_ffmpeg(bool) 在音频能用wave库输出时是否强制使用ffmpeg导出 默认为False
        rate(int) 码率 对应ffmpeg/avconc中"-ab"参数 默认为None
        ffmpeg_para(list) ffmpeg/avconc自定义参数 默认为None
    """
    input_bytes = input_transform(input_voice)
    pcm = silk_decode(input_bytes)

    if audio_format is None:
        if isinstance(output_voice, (os.PathLike, str)):
            audio_format = Path(output_voice).suffix[1:]
        else:
            raise ValueError("Pls tell me what audio format to use")
    if codec is None:
        codec = choose_decoder(audio_format)

    if codec == Codec.wave:
        audio = wav_decode(pcm)
    elif codec == Codec.libsndfile:
        metadata = kwargs.get("metadata")
        quality = kwargs.get("quality")
        subtype = kwargs.get("subtype")
        audio = sndfile_decode(pcm, audio_format, subtype, quality, metadata)
    elif codec == Codec.ffmpeg:
        rate = kwargs.get("rate")
        metadata = kwargs.get("metadata")
        ffmpeg_para = kwargs.get("ffmpeg_para")
        audio = ffmpeg_decode(pcm, audio_format, ffmpeg_para, rate, metadata)

    return output_transform(output_voice, audio)
